/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};


/*
 *  Boilerplate code for encoding and decoding data transmitted to / received from TI devices.
 */

var targetAccessErrorHandler = function(error) {
    if (error && error.message) {
        throw error.message;
    } else {
        throw error;
    }
};


/**
 * This code implements the IPacketCodec interface.
 *
 * @constructor
 * @implements gc.databind.IPacketCodec
 */
CustomCodec = function() {};

// use default implementations of connect,disconnect,encode and decode functions:
CustomCodec.prototype = new gc.databind.IPollingPacketCodec();

CustomCodec.prototype.writeCmdTimer = null;
CustomCodec.prototype.maxBitWidth = 16;
CustomCodec.prototype.readCmdSymbols = null;
CustomCodec.prototype.writeCmdSymbols = null;
CustomCodec.prototype.pendingWriteCmds = null;
CustomCodec.prototype.refreshListener = null;
CustomCodec.prototype.preRefreshListener = null;
CustomCodec.prototype.skipReadsUntilNextReadCmdFlagSet = false;
CustomCodec.prototype.preRefreshTimeInMs = 30;
CustomCodec.prototype.regBlocks = null;

CustomCodec.prototype.setWriteCmdFlag = function(writeCmdTimer) {
    writeCmdTimer.hTimeout = null;
    if (this.pendingWriteCmds) {
        this.skipReadsUntilNextReadCmdFlagSet = true;
        for (var i = 0; i < this.pendingWriteCmds.length; i++) {
            var writeCmdSymbol = this.pendingWriteCmds[i];
            gc.console.log("CustomCodec","Set " + writeCmdSymbol);
            gc.target.access.writeValue(writeCmdSymbol, 1).fail(targetAccessErrorHandler);
        }
        this.pendingWriteCmds = null;
    }
};

CustomCodec.prototype.setReadCmdFlag = function() {
    if (this.readCmdSymbols) {
        for (var i = 0; i < this.readCmdSymbols.length; i++) {
            var readCmdSymbol = this.readCmdSymbols[i];
            gc.console.log("CustomCodec","Set " + readCmdSymbol);
            gc.target.access.writeValue(readCmdSymbol, 1).fail(targetAccessErrorHandler);
        }
        this.skipReadsUntilNextReadCmdFlagSet = false;
    }
};

CustomCodec.prototype.writeValue = function(symbol, value) {
    gc.console.log("CustomCodec","CustomCodec.writeValue(" + symbol.regName + "," + value + ")");
    var structName = "";
    var writeCmdName = "";
    if (symbol.parentGroup) {
        if (symbol.parentGroup.structName) {
            structName = symbol.parentGroup.structName + ".";
        }
        if (symbol.parentGroup.structWriteCmdName && symbol.parentGroup.structWriteCmdName !== 'undefined') {
            writeCmdName = structName + symbol.parentGroup.structWriteCmdName;
        }
    }
    var regValue = +value;
    var varsObj = {};
    for (var i = 0; i < symbol.fields.length; i++) {
        var field = symbol.fields[i];
        var fieldName = field.name;
        varsObj[fieldName] = this.getBitfieldValue(field, regValue, symbol.size);
    }

    if (writeCmdName && (writeCmdName.length > 0)) {
        if (this.writeCmdTimer && this.writeCmdTimer.hTimeout) {
            window.clearTimeout(this.writeCmdTimer.hTimeout);
        }
        this.writeCmdTimer = {
            symbol: symbol,
            hTimeout: null
        };

        if (!this.pendingWriteCmds) {
            this.pendingWriteCmds = [];
        }

        if (this.pendingWriteCmds.indexOf(writeCmdName) < 0) {
            this.pendingWriteCmds.push(writeCmdName);
        }
        this.writeCmdTimer.hTimeout = window.setTimeout(this.setWriteCmdFlag.bind(this, this.writeCmdTimer), this.preRefreshTimeInMs);
    }
    return gc.target.access.writeValue(structName + symbol.regName, varsObj).fail(targetAccessErrorHandler);
};

CustomCodec.prototype.getBitfieldValue = function(bitfield, regValue, regWidth) {
    if (!bitfield) return "";
    if ((bitfield.attrs) && (+bitfield.attrs.isReserved)) return 0;
    var bitPos = bitfield.start;
    var shiftMask = 0;
    var width = bitfield.stop - bitfield.start + 1;
    for (var i = 0; i < width; i++) {
        shiftMask = (shiftMask << 1) + 1;
    }
    var bitfieldMask = shiftMask << bitPos;
    var result = (+regValue & bitfieldMask) >> bitPos;
    if (bitfield.widget && bitfield.widget.type === 'select') {
        for (var j = 0; j < bitfield.widget.options.length; j++) {
            if (bitfield.widget.options[j].value === result) {
                if (bitfield.widget.options[j].enum !== undefined) {
                    result = bitfield.widget.options[j].enum; // use the enum value to configure the bitfield with
                } else {
                    result = bitfield.widget.options[j].value << bitPos;
                }
                break;
            }
        }
    } else {
        // Motorware bitfields are given bit-shifted values: restore the shifted position
        if (width > 1) {
            result = result << bitPos;
        }
    }
    return result;
};

CustomCodec.prototype.setBitfieldValue = function(bitfield, regValue, regWidth, value) {
    if (!bitfield) return;
    var mask = (1 << regWidth) - 1;
    var bitPos = bitfield.start;
    var shiftMask = 0;
    var width = bitfield.stop - bitfield.start + 1;
    for (var i = 0; i < width; i++) {
        shiftMask = (shiftMask << 1) + 1;
    }
    if (bitfield.widget && bitfield.widget.type === 'select') {
        for (var j = 0; j < bitfield.widget.options.length; j++) {
            if (bitfield.widget.options[j].value === value >> bitPos) {
                value = bitfield.widget.options[j].value;
                break;
            }
        }
    } else {
        // convert weird values for boolean variables into 1 or 0
        if (width === 1) {
            if (value !== 0) {
                value = 1;
            }
        } else {
            // Motorware bitfields are given bit-shifted values: shift them back to start at b0
            value = value >> bitPos;
        }
    }
    var orMask = shiftMask << bitPos;
    var andMask = mask & ~(orMask);
    var newValue = (regValue & andMask) | ((value << bitPos) & orMask);
    return gc.utils.getHexString(newValue, width);
};

CustomCodec.prototype.readValue = function(symbol) {
    var structName = "";
    if (symbol.parentGroup && symbol.parentGroup.structName) {
        structName = symbol.parentGroup.structName + "."
    }
    var origValue = +symbol.value;
    if ((CustomCodec.prototype.pendingWriteCmds && CustomCodec.prototype.pendingWriteCmds.length > 0) || (CustomCodec.prototype.skipReadsUntilNextReadCmdFlagSet)) {
        gc.console.log("CustomCodec","CustomCodec.readValue deferred for " + symbol.regName + " due to pending write.");
        return Q(origValue);
    }
    var _self = this;
    gc.console.debug("CustomCodec","CustomCodec.readValue(" + symbol.regName + ")");
    return gc.target.access.readValue(structName + symbol.regName).then(function(result) {
        var retVal = result;
        if (symbol.fields && symbol.fields.length > 0) {
            // combine the bitfield values that were read into a single register value
            // for use by the register view, etc.
            var regValue = origValue;
            for (var i = 0; i < symbol.fields.length; i++) {
                var field = symbol.fields[i];
                var fieldName = field.name;
                if (result.hasOwnProperty(fieldName)) {
                    regValue = _self.setBitfieldValue(field, regValue, symbol.size, result[fieldName]);
                }
            }
            retVal = +regValue;
        }
        return retVal;
    }).fail(targetAccessErrorHandler);
};

CustomCodec.prototype.onPreRefresh = function(oldValue, newValue, progress) {
    if (this.pendingWriteCmds && this.pendingWriteCmds.length > 0) {
        gc.console.log("CustomCodec","skipping setting readCmd flag because of pending write(s)")
    } else {
        gc.console.debug("CustomCodec","preRefresh: set ReadCmd flag");
        this.setReadCmdFlag();
    }
};

CustomCodec.prototype.onRefresh = function(oldValue, newValue, progress) {
    gc.console.log("CustomCodec","refresh");
};

CustomCodec.prototype.connect = function(settings) {
    this.model = document.querySelector('ti-model-register');
    if (this.model) {
        var modelId = this.model.id;
        var refreshBind = gc.databind.registry.getBinding(modelId + '.$refresh_interval');
        if (refreshBind) {
            refreshBind.setPreRefreshInterval(this.preRefreshTimeInMs);
            this.preRefreshListener = new gc.databind.IChangedListener();
            this.preRefreshListener.onPreRefresh = this.onPreRefresh.bind(this);
            refreshBind.addPreRefreshListener(this.preRefreshListener);

            this.refreshListener = new gc.databind.IChangedListener();
            this.refreshListener.onRefresh = this.onRefresh.bind(this);
            refreshBind.addRefreshListener(this.refreshListener);

        }
        this.model.getRegisterViewInformation(0).then(function(deviceObj) {
            this.maxBitWidth = deviceObj.info.regsize;
            this.readCmdSymbols = [];
            this.regBlocks = deviceObj.regblocks;
            for (var i = 0; i < deviceObj.regblocks.length; i++) {
                var regBlock = deviceObj.regblocks[i];
                var structName = "";
                if (regBlock.structName && regBlock.structName.length > 0) {
                    structName = regBlock.structName + ".";
                }
                if (regBlock.structReadCmdName) {
                    var readCmdSymbol = structName + regBlock.structReadCmdName;
                    this.readCmdSymbols.push(readCmdSymbol);
                }
                if (regBlock.structWriteCmdName) {
                    var writeCmdSymbol = structName + regBlock.structWriteCmdName;
                    this.writeCmdSymbols.push(writeCmdSymbol);
                }
            }
        }.bind(this));
    }
    var regPage = document.querySelector('#ti_widget_register_page');
    if (regPage){
        regPage.preReadDelay = this.preRefreshTimeInMs;
        regPage.addEventListener('pre_read_reg', function(event) {
            this.setReadCmdFlag();
        }.bind(this));
    }

};

// registerCustomCodec has to be called right at startup in order for the codec to be recognized and used
gc.console.log("CustomCodec","CustomCodec: about to register...");
var baseCodecs;
gc.databind.registerCustomCodec("customRegModelCodec", CustomCodec, baseCodecs);

